//----------------------------------------------------------------------------
//  EDGE Main Lib Functions
//----------------------------------------------------------------------------
// 
//  Copyright (c) 2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------

#include "../i_defs.h"
#include "../l_epi.h"

typedef struct
{
  shutdown_f *f;
  void *data;
}
shutdown_routine_t;

static shutdown_routine_t *shutdown_seq = NULL;
static int num_shutdowns = 0;

/*

Startup and shutdown



*/

//
// L_Startup
//
void L_Startup(void)
{
  L_InitMemorySystem();

  I_SystemPreInit();

  L_InitMemoryHelpers();

  ENGINE_Main();
}

void L_RegisterShutdownRoutine(shutdown_f *f, void *data)
{
  L_Resize(shutdown_seq, shutdown_routine_t, num_shutdowns + 1);
  shutdown_seq[num_shutdowns].f = f;
  shutdown_seq[num_shutdowns].data = data;
  num_shutdowns++;
}

//
// L_Shutdown
//
void L_Shutdown(void)
{
  int i;
  
  for (i = num_shutdowns - 1; i >= 0; i--)
  {
    shutdown_seq[i].f(shutdown_seq[i].data);
  }

  L_Free(shutdown_seq);

  L_ShutdownMemorySystem();

  I_CloseProgram(0);
}

#define MSGBUFSIZE 4096

static text_receiver_f **receivers = NULL;
static int num_receivers = 0;

/*

Printf system
When

*/

void L_AddPrintReceiver(text_receiver_f *f)
{
  L_Resize(receivers, text_receiver_f, num_receivers + 1);
  receivers[num_receivers] = f;
  num_receivers++;
}

void L_RemovePrintReceiver(text_receiver_f *f)
{
  int i;
  
  for (i = 0; i < num_receivers; i++)
    if (f == receivers[i])
      break;
      
  DEV_ASSERT2(i < num_receivers);
  for (; i < num_receivers - 1; i++)
    receivers[i] = receivers[i + 1];

  num_receivers--;
  Z_Resize(receivers, text_receiver_f*, num_receivers);
}

void L_PrintString(const char *s)
{
  int i;
  
  for (i = 0; i < num_receivers; i++)
  {
    receivers[i](s);
  }
}

//
// L_VPrintf
//
void L_VPrintf(const char *format, va_list argptr)
{
  char printbuf[MSGBUFSIZE];

  printbuf[MSGBUFSIZE-1] = 0;

  // Print the message into a text string
  vsprintf(printbuf, format, argptr);

  // I hope nobody is printing strings longer than 4096 chars...
  DEV_ASSERT2(printbuf[MSGBUFSIZE-1] == 0);

  L_PrintString(printbuf);
}

//
// L_Printf
//
// Converts the message to a plain string, and sends it to each of its receivers.
void L_Printf(const char *message, ...)
{
  va_list argptr;

  va_start(argptr, message);
  L_VPrintf(message, argptr);
  va_end(argptr);
}

//
// L_Error
//
void L_Error(const char *error, ...)
{
  va_list argptr;
  // -ES- 2000/01/07 Avoid infinite recursion.
  static boolean_t firsttime = true;

  if (firsttime)
    firsttime = false;
  else
    return;

  // Bring the system to a close
  I_SystemShutdown();

  // Message last, so it actually prints on the screen
  va_start(argptr, error);
  if (debugfile)
  {
    fprintf(debugfile, "\n");
    vfprintf(debugfile, error, argptr);
    fflush(debugfile);
  }
  fprintf(stdout, "\n");
  vfprintf(stdout, error, argptr);
  fprintf(stdout, "\n");

  va_end(argptr);
  fflush(stdout);

#ifdef DEVELOPERS
  if (traceback_on_error)
    I_TraceBack();
#endif

  I_CloseProgram(-1);
}
