//----------------------------------------------------------------------------
//  EDGE Memory Allocation Primitives
//----------------------------------------------------------------------------
// 
//  Copyright (c) 2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------

#include "../i_defs.h"
#include "../l_epi.h"

#define ZONEID  0x1d4a11f1

static cache_flusher_f **cache_flushers = NULL;
static int num_flushers = 0;

#ifdef DEVELOPERS
typedef struct mallocheader_s
{
  int size;
} mallocheader_t;

#define CHECK_PTR(h) DEV_ASSERT((*(int *)((char *)(h) + (h)->size + sizeof(mallocheader_t)) == ZONEID), ("Block without ZONEID"))
#endif

//
// L_RegisterCacheFlusher
//
// Tells the memory system that f can be called to free up some memory.
//
void L_RegisterCacheFlusher(cache_flusher_f *f)
{
  L_Resize(cache_flushers, cache_flusher_f *, num_flushers + 1);
  cache_flushers[num_flushers] = f;
  num_flushers++;
}

void L_UnregisterCacheFlusher(cache_flusher_f *f)
{
  int i;

  // find it
  for (i = 0; i < num_flushers; i++)
    if (f == cache_flushers[i])
      break;

  if (i == num_flushers)
    I_Error("L_UnregisterCacheFlusher: %p is not a registered flusher!", f);
  memmove(&cache_flushers[i], &cache_flushers[i+1], sizeof(cache_flusher_f) * (num_flushers - i - 1));
}

//
// FlushCaches
//
// Calls other parts of the code to trim any "cached" data, like
// unused structures stored on an quick-alloc list.  It gets called
// when the zone cannot fulfill a L_Malloc request, in the hope that
// after flushing the memory will be available.
//
// `urge' is one of the L_Urgency* values, and the more urgent the
// request is, the harder that the cache flushing code should try to
// free memory.  For example, sprites & textures can be reconstructed
// (albeit slowly) from info in the WAD file(s), and in the worst case
// scenario (urge == L_UrgencyExtreme) the memory could be freed.
//
// -AJA- 1999/09/16: written.
// -ES- 1999/12/18 Written.

static void FlushCaches(flusher_urgency_e urge, int size)
{
  int i;

  // Call all cache flusher at the current urgency level
  for (i = 0; i < num_flushers; i++)
    (*cache_flushers[i])(urge);

#ifdef DEVELOPERS
  {
    // Output flush count.
    static int count = 0;
    L_WriteDebug("Flush %d", count++);
  }
#endif
}

//
// L_Free
//
void L_Free(void *ptr)
{
  // -ES- FIXME: Decide whether we should allow L_Free(NULL) and L_Malloc(0)
  if (ptr == NULL)
    return;
#ifdef DEVELOPERS
  {
    mallocheader_t *h = (mallocheader_t *)ptr - 1;
    CHECK_PTR(h);
#ifdef INTOLERANT_MATH
    // -ES- 2000/04/08 Trash all data.
    memset(h, -1, h->size + sizeof(mallocheader_t) + sizeof(int));
#endif
    free(h);
  }
#else
  free(ptr);
#endif
}

//
// L_Malloc
//
// Memory allocation.
//
void *L_Malloc2(int size)
{
#ifdef DEVELOPERS
  mallocheader_t *p;
#else
  void *p;
#endif
  int allocsize;
  flusher_urgency_e flush_urge = 0;

  if (size == 0)
    return NULL;

#ifdef DEVELOPERS
  allocsize = sizeof(mallocheader_t) + size + sizeof(int);
#else
  allocsize = size;
#endif

  while (NULL == (p = malloc(allocsize)))
  {
    if (flush_urge == L_UrgencyExtreme)
      I_Error("L_Malloc: failed on allocation of %i bytes", size);
    flush_urge++;
    FlushCaches(flush_urge, allocsize);
  }
#ifdef DEVELOPERS
  p->size = size;
  p++;
  *(int *)((char *)p + size) = ZONEID;
#ifdef INTOLERANT_MATH
  // -ES- 2000/03/28 Will turn all floats into -NaN, making uninitialised
  // elements easier to find.
  memset(p, -1, size);
#endif
#endif
  return p;
}

//
// L_ReMalloc
//
// Memory reallocation.
//
void *L_ReMalloc2(void *ptr, int size)
{
#if defined INTOLERANT_MATH && defined DEVELOPERS
  // The intolerant remalloc will always trash the old memory block with -1:s,
  // and also trash anything undefined in the new block.
  void *newp;
  mallocheader_t *h;

  if (size == 0)
  {
    if (ptr)
      L_Free(ptr);
    return NULL;
  }
  
  if (ptr == NULL)
    return L_Malloc(size);

  // All trashing of undefined data will be done by L_Malloc/L_Free.
  newp = L_Malloc(size);
  h = (mallocheader_t *)ptr - 1;
  CHECK_PTR(h);
  I_MoveData(newp, ptr, MIN(size, h->size));
  L_Free(ptr);
  
  return newp;
#else

#ifdef DEVELOPERS
  mallocheader_t *h, *newp;
#else
  void *h, *newp;
#endif
  int allocsize;
  flusher_urgency_e flush_urge = 0;

  if (size == 0)
  {
    if (ptr)
      L_Free(ptr);

    return NULL;
  }

#ifdef DEVELOPERS
  if (ptr == NULL)
  {
    h = NULL;
  }
  else
  {
    h = (mallocheader_t *)ptr - 1;
    CHECK_PTR(h);
  }
  allocsize = size + sizeof(mallocheader_t) + sizeof(int);
#else
  h = ptr;
  allocsize = size;
#endif

  while (NULL == (newp = realloc(h, allocsize)))
  {
    if (flush_urge == L_UrgencyExtreme)
      I_Error("L_ReMalloc: failed on allocation of %i bytes", size);
    flush_urge++;
    FlushCaches(flush_urge, allocsize);
  }

#ifdef DEVELOPERS
  newp->size = size;
  newp++;
  *(int *)((char *)newp + size) = ZONEID;
#endif
  return newp;
#endif
}

//
// L_Calloc
//
// Allocates memory and clears it to 0.
//
void *L_Calloc2(int size)
{
  byte *data = (byte *) L_Malloc(size);
  L_Clear(data, byte, size);
  return (void *) data;
}

//
// L_CheckHeap
//
void L_CheckHeap(void)
{
}

//
// Just make sure that everything has been shut down properly
//
void L_ShutdownMemorySystem(void)
{
#ifdef LEAK_HUNT
  DEV_ASSERT(num_flushers == 0, ("INTERNAL ERROR: ShutdownMemoryAlloc: All flushers have not been shut down!"))
#endif
}

//
// L_InitMemoryAlloc
//
void L_InitMemorySystem(void)
{
  // Nothing needs to be initialised
}
