//----------------------------------------------------------------------------
//  EDGE High-level Memory Helper Functions
//----------------------------------------------------------------------------
// 
//  Copyright (c) 2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------

#include "../i_defs.h"
#include "../l_epi.h"

static stack_array_t *stack_arrays = NULL;

//
// L_StrDup
//
// Duplicates the string.
char *L_StrDup(const char *s)
{
  int size;
  char *ret;

  if (s == NULL)
    return NULL;

  size = strlen(s) + 1;

  ret = L_New(char, size);

  L_MoveMem(ret, s, char, size);

  return ret;
}

//
// L_ClearStackArray
//
// Deletes the contents of the growable array, but doesn't free it.
void L_ClearStackArray(stack_array_t *a)
{
  int i;
  void **p = *a->ptr;
  stack_array_t *prev;

#ifdef DEVELOPERS
  if (a->locked)
    I_Error("L_ClearStackArray: Locked!");
#endif

  // kill all the elements
  if (a->alloc_bunch > 0)
  {
    for (i = 0; i < a->max; i += a->alloc_bunch)
    {
      L_Free(p[i]);
    }
  }

  // kill the array itself
  L_Free(p);

  // kill the pointer reference
  *a->ptr = NULL;
  // kill this ref too: A stack array is defined to be uninitialised iff
  // ptr is null.
  a->ptr = NULL;

  // unlink it from the list.
  if (a == stack_arrays)
    // special case if we were first in the list
    stack_arrays = a->next;
  else
  {
    // search for the element before a, and change its pointer
    for (prev = stack_arrays; prev->next != a; prev = prev->next)
      DEV_ASSERT(prev->next, ("L_ClearStackArray: List is broken"));

    prev->next = a->next;
  }
}

//
// L_DeleteStackArray
//
// Kills & frees a growable array
void L_DeleteStackArray(stack_array_t *a)
{
  L_ClearStackArray(a);
  L_Free(a);
}

// If you for some reason must set a pointer directly to the address of
// one of the elements of *ptr, you must first lock the array. This means
// that it won't be flushed by the cache code, and that it can't
// increase its size, so *ptr will never need to be reallocated.
// You must remember to unlock the stack array, otherwise the next
// L_SetArraySize will probably fail.
void L_LockStackArray(stack_array_t *a)
{
  a->locked++;
}

void L_UnlockStackArray(stack_array_t *a)
{
#ifdef DEVELOPERS
  if (a->locked == 0)
    I_Error("L_UnlockStackArray: Already unlocked!");
#endif
  a->locked--;
}

//
// L_InitStackArray
//
// Initialises the growable array.
// A growable array is a type that can be used for dynamic arrays, whose sizes
// will vary. With a single function call, you will be able to tell how much
// of the array you need, and it will see that there are that many elements,
// without moving any of the existing elements (and thereby spoiling any
// pointers to them). It will also reduce the number of memory allocation
// calls and memory blocks.
// ptr: Points to the array pointer.
// elem_size: The size of each element.
// alloc_bunch: Number of elements to allocate at a time. 0 for default value.

void L_InitStackArray(stack_array_t *a, void ***ptr, int elem_size, int alloc_bunch)
{
  // a->ptr must always be set to NULL. The point is that it is possible
  // to do this kind of checks.
  DEV_ASSERT(a->ptr == NULL, ("L_InitStackArray: Array already initialised"));

  a->elem_size = elem_size;

  if (alloc_bunch == 0)
  {
    // Assuming that about 16 bytes are wasted in some way per allocated
    // block, the default is to let no more than 1/16 of the memory be wasted
    // in that way.
    alloc_bunch = 1 + 16 * 16 / elem_size;
  }

  a->alloc_bunch = alloc_bunch;

  a->ptr = ptr;
  *ptr = NULL;
  // these will be initialised later.
  a->num = a->max = 0;
  a->locked = 0;

  // insert array into the head of the list.
  a->next = stack_arrays;
  stack_arrays = a;
}

//
// L_CreateStackArray
//
// Allocates and initialises a growable array.
stack_array_t *L_CreateStackArray(void ***ptr, int elem_size, int alloc_bunch)
{
  stack_array_t *a;

  a = L_New(stack_array_t, 1);
  a->ptr = NULL;
  L_InitStackArray(a, ptr, elem_size, alloc_bunch);
  return a;
}

//
// Does the actual array reallocations.
//
static void SetActualArraySize(stack_array_t *a, int num)
{
  int newmax;
  int i;
  byte *elem;

#ifdef DEVELOPERS
  if (a->locked)
    I_Error("SetActualArraySize: Locked!");
#endif

  // alloc_bunch < 0 means that it is an array of elements, not
  // an array of element pointers. The array will grow -alloc_bunch
  // elements at a time.
  if (a->alloc_bunch < 0)
  {
    *a->ptr = L_ReMalloc(*a->ptr, a->elem_size * num);
    a->max = num;
  }
  else
  {
    newmax = num + a->alloc_bunch - 1;
    newmax -= newmax % a->alloc_bunch;
    // if we are shrinking the array, just free the unneeded blocks.
    while (num <= a->max - a->alloc_bunch)
    {
      a->max -= a->alloc_bunch;
      L_Free((*a->ptr)[a->max]);
    }
    L_Resize(*a->ptr, void *, newmax);
    // allocate one bunch at a time, and init elements.
    while (num > a->max)
    {
      elem = (byte *) L_Malloc(a->alloc_bunch * a->elem_size);

      for (i = 0; i < a->alloc_bunch; i++, elem += a->elem_size)
        (*a->ptr)[a->max+i] = elem;

      a->max += a->alloc_bunch;
    }
#ifdef DEVELOPERS
    // should never happen...
    if (newmax != a->max)
      I_Error("L_SetArraySize: Internal Error");
#endif
  }
}

//
// L_SetArraySize
//
// Sets the number of used elements in a to num, allocating extra space if
// needed.
void L_SetArraySize(stack_array_t *a, int num)
{
#ifdef DEVELOPERS
  if (a->locked && num > a->num)
    I_Error("L_SetArraySize: Locked!");
#endif
  // we are done if the new size is below the maximum.
  if (num > a->max)
  {
    if (a->alloc_bunch < 0)
      // allocate some extra elements
      SetActualArraySize(a, num - a->alloc_bunch);
    else
      // extra allocation will be done automatically
      SetActualArraySize(a, num);
  }
  a->num = num;
}

//
// Frees all unused elements of the arrays.
//
static void FlushStackArrays(flusher_urgency_e urge)
{
  stack_array_t *a;

  for (a = stack_arrays; a; a = a->next)
  {
    if (!a->locked)
      SetActualArraySize(a, a->num);
  }
}

//
// Not much has to be done here, actually
//
static void ShutdownMemoryHelpers(void *data)
{
#ifdef LEAK_HUNT
  DEV_ASSERT(stack_arrays == NULL, ("INTERNAL ERROR: ShutdownMemoryHelpers: All stack_arrays have not been destroyed!"));
#endif
  L_UnregisterCacheFlusher(FlushStackArrays);
}

//
// L_InitMemoryHelpers
//
void L_InitMemoryHelpers(void)
{
  // Register stack array flusher. That system doesn't need any
  // further initialisation.
  L_RegisterCacheFlusher(FlushStackArrays);
  L_RegisterShutdownRoutine(ShutdownMemoryHelpers, NULL);
}

